# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from __future__ import annotations

import inspect
import os
import sys
import warnings
from pathlib import Path
from typing import Any

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Pyqrlew'
copyright = '2023, Sarus Technologies'
author = 'Sarus Technologies'
github_root = "https://github.com/Qrlew/pyqrlew"
git_branch = 'main'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # 'myst_parser',
    'myst_nb',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon', # it understands google and numpy style doc strings
    'sphinx.ext.intersphinx', # it add links to objects
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinxcontrib.bibtex',
    "sphinx.ext.linkcode",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Extension settings  -----------------------------------------------------

# sphinx.ext.intersphinx - link to other projects' documentation
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Sphinx-copybutton - add copy button to code blocks
# https://sphinx-copybutton.readthedocs.io/en/latest/index.html
# strip the '>>>' and '...' prompt/continuation prefixes.
copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True

# Enable numref
numfig = True
bibtex_bibfiles = ['qrlew.bib']

# MyST conf
myst_enable_extensions = [
    'dollarmath',
    'amsmath',
]
myst_dmath_allow_labels = True
myst_dmath_double_inline = True


autodoc_typehints = 'both' # type hints both in the signature and in the docstring 
# autodoc_typehints_description_target = 'documented'
autodoc_docstring_signature = False
autodoc_member_order = 'bysource' # show members by their source order (it doesn't seem to have an effect)
# Controls notebook execution
nb_execution_mode = 'off'#'off'/'cache'


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'

html_static_path = ['_static']

html_theme_options = {
    'navbar_start': ['navbar-logo'],
    'navbar_center': ['navbar-nav'],
    'navbar_end': ['navbar-icon-links'],
    'navbar_persistent': ['search-button'],
    'icon_links': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/Qrlew',
            'icon': 'fa-brands fa-github',
            'type': 'fontawesome',
        },
        {
            'name': 'Web Page',
            'url': 'https://qrlew.github.io/',
            'icon': 'fa-solid fa-house',
            'type': 'fontawesome',
        },
        {
            'name': 'Discord',
            'url': 'https://discord.com/invite/XQ9jYshj2p',
            'icon': 'fa-brands fa-discord',
            'type': 'fontawesome',
        },
        {
            'name': 'X',
            'url': 'https://x.com/sarus_tech',
            'icon': 'fa-brands fa-twitter',
            # The default for `type` is `fontawesome` so it is not actually required in any of the above examples as it is shown here
        },
    ],
}

import pyqrlew

# sphinx-ext-linkcode - Add external links to source code
# https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html
def linkcode_resolve(domain: str, info: dict[str, Any]) -> str | None:
    """
    Determine the URL corresponding to Python object.

    Based on pandas equivalent:
    https://github.com/pandas-dev/pandas/blob/main/doc/source/conf.py#L629-L686
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            with warnings.catch_warnings():
                # Accessing deprecated objects will generate noisy warnings
                warnings.simplefilter("ignore", FutureWarning)
                obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        try:  # property
            fn = inspect.getsourcefile(inspect.unwrap(obj.fget))
        except (AttributeError, TypeError):
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except TypeError:
        try:  # property
            source, lineno = inspect.getsourcelines(obj.fget)
        except (AttributeError, TypeError):
            lineno = None
    except OSError:
        lineno = None

    linespec = f"#L{lineno}-L{lineno + len(source) - 1}" if lineno else ""

    # https://github.com/Qrlew/pyqrlew/blob/main/pyqrlew/python/pyqrlew/io/postgresql.py#L16-L167
    # https://github.com/Qrlew/pyqrlew/blob/main/python/pyqrlew/io/postgresql.py
    # fn is the absolute path of a specific source in the environment where pyqrlew is installed
    # to get the relative path I could use pyqrlew.__file__ as follow 
    fn = Path("python") / Path("pyqrlew") /  os.path.relpath(fn, start=os.path.dirname(pyqrlew.__file__))
    return f"{github_root}/blob/{git_branch}/{fn}{linespec}"



from __future__ import annotations
import typing as t
import enum

from pyqrlew.typing import PrivacyUnit, SyntheticData

class _Dataset:
    """Class that ...."""
    def __new__(self, dataset: str, schema: str, size: str) -> '_Dataset': ...
    @property
    def schema(self) -> str: ...
    @property
    def size(self) -> t.Optional[str]: ...
    def with_range(self, schema_name: str, table_name: str, field_name: str, min: float, max: float) -> '_Dataset': ...
    def with_possible_values(self, schema_name: str, table_name: str, field_name: str, possible_values: t.Iterable[str]) -> '_Dataset': ...
    def with_constraint(self, schema_name: str, table_name: str, field_name: str, constraint: t.Optional[str]) -> '_Dataset': ...
    def relations(self) -> t.Iterable[t.Tuple[t.List[str], '_Relation']]: ...
    def relation(self, query: str, dialect: t.Optional['Dialect']) -> '_Relation': ...
    def from_queries(self, queries: t.Iterable[t.Tuple[t.Iterable[str], str]], dialect: t.Optional['Dialect']) -> '_Dataset': ...
    def __str__(self) -> str: ...


class _Relation:
    """Class that... """
    @staticmethod
    def from_query(query: str, dataset: _Dataset, dialect: t.Optional['Dialect']) -> '_Relation': ...
    def __str__(self) -> str: ...
    def dot(self) -> str: ...
    def schema(self) -> str: ...
    def type_(self) -> str: ...
    def to_query(self, dialect: t.Optional['Dialect']=None) -> str: ...
    def rewrite_as_privacy_unit_preserving(
        self,
        dataset: _Dataset,
        privacy_unit: PrivacyUnit,
        epsilon_delta: t.Dict[str, float],
        max_multiplicity: t.Optional[float]=None,
        max_multiplicity_share: t.Optional[float]=None,
        synthetic_data: t.Optional[SyntheticData]=None,
        strategy: t.Optional['Strategy']=None
    ) -> '_RelationWithDpEvent': ...
    def rewrite_with_differential_privacy(
        self,
        dataset: _Dataset,
        privacy_unit: PrivacyUnit,
        epsilon_delta: t.Dict[str, float],
        max_multiplicity: t.Optional[float]=None,
        max_multiplicity_share: t.Optional[float]=None,
        synthetic_data: t.Optional[SyntheticData]=None
    ) -> '_RelationWithDpEvent': ...
    def with_field(self, name: str, expr: str) -> '_Relation': ...
    def rename_fields(self, fields: t.Iterable[t.Tuple[str, str]])  -> '_Relation': ...
    def compose(self, relations: t.Iterable[t.Tuple[t.Iterable[str], '_Relation']]) -> '_Relation': ...

class _RelationWithDpEvent:
    def relation(self) -> _Relation: ...
    def dp_event(self) -> DpEvent: ...


class DpEvent:
    def to_dict(self) -> t.Mapping[str, t.Union[str, float]]: ...
    def to_named_tuple(self) -> t.NamedTuple: ...


class Dialect(enum.Enum):
    """Supported dialects"""
    PostgreSql=1
    MsSql=2
    BigQuery=3

class Strategy(enum.Enum):
    Soft=1
    Hard=2


import numpy as np
from typing import List, Callable, Union
import typing as t
import textwrap
from .stochatic_dataset import StochasticDatabase
from .tester import StochasticTester
import pandas as pd # type: ignore

# Optionally import matplotlib
try:
    import matplotlib.pylab as plt  # type: ignore  # module is installed, but missing library stubs or py.typed marker 
    plt.rcParams['text.usetex'] = True
except ImportError:
    plt = None

def divergence(array1: np.ndarray, array2: np.ndarray, epsilon: float) -> float:
    arr = [max(l1 - np.exp(epsilon)*l2, 0) for (l1, l2) in zip(array1, array2)]
    return float(np.sum(arr))

def privacy_profile(ref_results: np.ndarray, adj_results: List[np.ndarray], epsilon: float) -> float:
    divergences = [
        divergence(ref_results, adj_res, epsilon)
        for adj_res in adj_results
    ]
    return float(max(divergences))

def compute_distribution(my_list: list, bins: Union[int, List[int]]) -> np.ndarray:
    values, _ = np.histogram(my_list, bins=bins)
    values = values / np.sum(values)
    return values

def plot_utility(dp_results: dict, true_res: float) -> None:
    if plt is None:
        raise ModuleNotFoundError
    [
        plt.hist(results, bins=20, alpha=0.5, label = fr"$\varepsilon = {epsilon}$")
        for (epsilon, results) in dp_results.items()
    ]
    plt.axvline(true_res, color = 'red')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.legend()

def plot_adjacent_histograms(array1: List[float], array2: List[float], epsilon:float, delta: float, nbins:float=20, alpha:float=0.5, color1:str='blue', color2:str='orange') -> None:
    if plt is None:
        raise ModuleNotFoundError
    xmin = max(min(array1), min(array2))
    xmax = min(max(array1), max(array2))
    bins = np.linspace(xmin, xmax, int(nbins))
    values1 = compute_distribution(array1, t.cast(List[int], bins))
    values2 = compute_distribution(array2, t.cast(List[int], bins))
    x = bins[1:] + (bins[1] -bins[0])

    plt.bar(x, np.exp(epsilon) * values2 + delta, width=bins[1]-bins[0], alpha=alpha, color=color2, label=r'$e^{\varepsilon}P(f(\mathcal{D}_2)) + \delta$')
    plt.bar(x, values1, alpha=alpha, color=color1, width=bins[1]-bins[0], label=r'$P(f(\mathcal{D}_1))$')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.legend()

def plot_privacy_profile(array: List[float], arrays: List[List[float]], nbins:int=20, epsilons: List[float]=list(np.linspace(0, 3, 100))) -> None:
    if plt is None:
        raise ModuleNotFoundError
    xmin = max(min([min(a) for a in arrays]), min(array))
    xmax = min(max([max(a) for a in arrays]), max(array))
    bins = int(np.linspace(xmin, xmax, nbins))

    ref_distribution = compute_distribution(array, bins)
    adj_distributions = [
        compute_distribution(res, bins)
        for res in arrays
    ]
    pprofiles = [
        privacy_profile(ref_distribution, adj_distributions, epsilon) for epsilon in epsilons
    ]
    plt.plot(epsilons, pprofiles, '-', color='navy')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.xlabel(r'$\varepsilon$')
    plt.ylabel(r'$\delta$')

def save_data(filename: str, arr_ref: List[float], arr_adj: List[List[float]]) -> None:
    df = pd.DataFrame()
    df['ref'] = arr_ref
    for i, adj in enumerate(arr_adj):
        df[f'adj_{i}'] = adj
    df.to_csv(filename, index=False)
    print(f"file {filename} written.")

def run_test(database: StochasticDatabase, stochastic_tester: StochasticTester, query: str, adj_query_func: Callable[[int], str], name: str, n_sim: int)-> None:
    if plt is None:
        raise ModuleNotFoundError
    print(f"Query = {query}")
    dp_results = stochastic_tester.utility_per_epsilon(query, epsilons=[1., 5., 10.], n_sim=n_sim)
    true_res = database.execute(query)[0][0]
    plt.figure(figsize=(6, 15))
    plt.subplot(3, 1, 1)
    plot_utility(dp_results=dp_results, true_res=true_res)

    epsilon = 1.
    ref_res = stochastic_tester.compute_results(query, epsilon=epsilon, n_sim=n_sim)
    adj_res = [
        stochastic_tester.compute_results(
            adj_query_func(int(i)),
            epsilon=epsilon,
            n_sim=n_sim
        )
        for i in [np.random.uniform(0, 100) for _ in range(10)]
    ]
    save_data(f"{name}.txt", ref_res, adj_res)

    nbins = int(n_sim / 200)
    plt.subplot(3, 1, 2)
    plot_adjacent_histograms(ref_res, adj_res[0], epsilon=epsilon, delta=stochastic_tester.delta, nbins=nbins)

    plt.subplot(3, 1, 3)
    plot_privacy_profile(ref_res, adj_res, nbins=nbins)
    plt.axhline(stochastic_tester.delta, color='r')

    wrapped_query = textwrap.fill(query, width=30)  # Adjust the width as needed
    plt.suptitle(wrapped_query, fontsize=12)
    plt.tight_layout()
    plt.savefig(f"{name}.png")
    print("================================")



from pyqrlew.io import PostgreSQL
from pyqrlew import Relation, Dialect, Strategy
from pyqrlew.utils import display_graph
from pyqrlew.wrappers import Dataset
import pytest


@pytest.fixture
def extract_dataset():
    database = PostgreSQL()
    dataset = database.extract()
    return dataset


def test_from_query(extract_dataset):
    query = "SELECT * FROM extract.census"
    _ = Relation.from_query(query, extract_dataset)
    _ = Relation.from_query(query, extract_dataset, dialect=Dialect.PostgreSql)
    _ = Relation.from_query(query, extract_dataset, dialect=Dialect.BigQuery)
    _ = Relation.from_query(query, extract_dataset, dialect=Dialect.MsSql)
    

def test_rewrite_as_privacy_unit_preserving_soft(extract_dataset):
    query = "SELECT * FROM extract.census"
    rel = Relation.from_query(query, extract_dataset)
    privacy_unit = [
        ("census", [], "_PRIVACY_UNIT_ROW_")
    ]
    epsilon_delta={"epsilon": 1.0, "delta": 1e-3}
    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
    )
    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
        strategy=Strategy.Soft
    )

    privacy_unit = ([
        ("census", [], "age")
    ], False)

    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
        strategy=Strategy.Soft
    )

    privacy_unit = ([
        ("census", [], "age", "education_num")
    ], False)

    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
        strategy=Strategy.Soft
    )

def test_rewrite_as_privacy_unit_preserving_hard(extract_dataset):
    query = "SELECT age, COUNT(*) FROM extract.census GROUP BY age"
    rel = Relation.from_query(query, extract_dataset)
    privacy_unit = [
        ("census", [], "_PRIVACY_UNIT_ROW_")
    ]
    epsilon_delta={"epsilon": 1.0, "delta": 1e-3}
    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
    )
    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
        strategy=Strategy.Hard
    )

    privacy_unit = ([
        ("census", [], "age")
    ], False)

    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
        strategy=Strategy.Hard
    )

    privacy_unit = ([
        ("census", [], "age", "education_num")
    ], False)

    _ = rel.rewrite_as_privacy_unit_preserving(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
        strategy=Strategy.Hard
    )

    with pytest.raises(RuntimeError) as e_info:
        _ = rel.rewrite_as_privacy_unit_preserving(
            dataset=extract_dataset,
            synthetic_data=None,
            privacy_unit=privacy_unit,
            epsilon_delta=epsilon_delta,
            strategy=Strategy.Soft
        )
        assert e_info == "UnreachableProperty: privacy_unit_preserving is unreachable"


def test_rewrite_with_differential_privacy(extract_dataset):
    query = "SELECT age, COUNT(*) FROM extract.census GROUP BY age"
    rel = Relation.from_query(query, extract_dataset)
    privacy_unit = [
        ("census", [], "_PRIVACY_UNIT_ROW_")
    ]
    epsilon_delta={"epsilon": 1.0, "delta": 1e-3}
    _ = rel.rewrite_with_differential_privacy(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
    )

    privacy_unit = ([
        ("census", [], "age")
    ], False)

    _ = rel.rewrite_with_differential_privacy(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
    )

    privacy_unit = ([
        ("census", [], "age", "education_num")
    ], False)

    _ = rel.rewrite_with_differential_privacy(
        dataset=extract_dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
    )

def test_rename_fields(extract_dataset):
    query = "SELECT age AS age, COUNT(*) AS count_all FROM extract.census GROUP BY age"
    rel = Relation.from_query(query, extract_dataset)
    fields = [("age", "renamed_age")]
    rel = rel.rename_fields(fields)
    assert rel.schema() == '{renamed_age: int[20 90] (UNIQUE), count_all: int[0 199]}'
    # display_graph(rel.dot())


def test_compose(extract_dataset):
    queries = [
        (("dataset_name", "my_schema", "boomers",), "SELECT * FROM extract.census WHERE age >= 60"),
        (("dataset_name", "my_schema", "genx",), "SELECT * FROM extract.census WHERE age >= 40 AND age < 60"),
        (("dataset_name", "my_schema", "millenials",), "SELECT * FROM extract.census WHERE age >= 30 AND age < 40"),
        (("dataset_name", "my_schema", "genz",), "SELECT * FROM extract.census WHERE age < 30"),
    ]
    relations = [
        (path, Relation.from_query(query, extract_dataset)) for (path, query) in queries
    ]

    new_ds = extract_dataset.from_queries(queries)
    new_query = """SELECT age, marital_status, COUNT(*) AS count_all
        FROM genx
        GROUP BY age, marital_status
        ORDER BY age, marital_status 
    """
    rel = Relation.from_query(new_query, new_ds)
    # display_graph(rel.dot())
    composed = rel.compose(relations)
    assert composed.schema() == '{age: int[40 60], marital_status: str{Divorced, Married-civ-spouse, Married-spouse-absent, Never-married, None, Separated, Widowed}, count_all: int[0 199]}'
    # display_graph(composed.dot())

def test_with_filed(extract_dataset):
    query = "SELECT age AS age, COUNT(*) AS count_all FROM extract.census GROUP BY age"
    rel = Relation.from_query(query, extract_dataset)
    new_field_name = "custom_field"
    new_field_expr = "false"
    new_rel = rel.with_field(name=new_field_name, expr=new_field_expr)
    assert new_rel.schema() == '{custom_field: bool{false}, age: int[20 90] (UNIQUE), count_all: int[0 199]}'

    new_field_name = "custom_null_field"
    new_field_expr = "NULL"
    new_rel = new_rel.with_field(name=new_field_name, expr=new_field_expr)
    assert new_rel.schema() == '{custom_null_field: option(any), custom_field: bool{false}, age: int[20 90] (UNIQUE), count_all: int[0 199]}'
    # display_graph(new_rel.dot())

def test_extra_queries(tables, engine):
    ds = Dataset.from_database("my_db", engine, schema_name=None)
    print(ds.relations())
    QUERIES = [
        ('SELECT "integer" FROM primary_table, primary_public_table'),
        (
            'SELECT "integer" FROM primary_table '
            "JOIN (SELECT * FROM primary_public_table) AS subq USING(id)"
        ),
        (
            """WITH tab1 AS (
                SELECT "integer" FROM primary_table
                JOIN primary_public_table USING(id)
            ),
            tab2 AS (
                SELECT * FROM primary_table JOIN secondary_table USING(id)
            )
            SELECT * FROM tab1
            """
        ),
        "SELECT t1.id FROM primary_public_table AS t1 JOIN secondary_public_table AS t2 ON(t1.id=t2.id)",
        """
        WITH "join_nagp" ("field_ljc8", "field_xkq4", "field_rv0k", "field_xn3k", "field_3z4q", "field_o2t5", "field_xfrs", "field_pqun", "field_90w5", "field_9tue") AS (
            SELECT * FROM "primary_public_table" AS "_LEFT_" JOIN "secondary_public_table" AS "_RIGHT_" ON ("_LEFT_"."id") = ("_RIGHT_"."id")
        ), "map_w9s_" ("_PRIVACY_UNIT_WEIGHT_", "_PRIVACY_UNIT_", "sarus_is_public", "id") AS (
            SELECT CAST(1 AS INTEGER) AS "_PRIVACY_UNIT_WEIGHT_", CAST(CASE WHEN (1) > (2) THEN 1 ELSE NULL END AS TEXT) AS "_PRIVACY_UNIT_", CAST(0 AS BOOLEAN) AS "sarus_is_public", "field_ljc8" AS "id"
            FROM "join_nagp"
        ) SELECT * FROM "map_w9s_"
        """,
    ]

    for query in QUERIES:
        rel = Relation.from_query(query, ds)
        query = rel.to_query()
        print(query)
        # display_graph(rel.dot())


def test_writing_and_rewriting_many_times(tables, engine):
    import sqlalchemy as sa
    ds = Dataset.from_database("my_db", engine, schema_name=None)
    query = """SELECT * FROM primary_table t1 JOIN primary_public_table t2 ON (t1.id=t2.id)"""

    with engine.connect() as conn:
        conn.execute(sa.text(query))

    for i in range(2):
        rel = Relation.from_query(query, ds)
        query = rel.to_query()
        print(query)
        with engine.connect() as conn:
            conn.execute(sa.text(query))
        # display_graph(rel.dot())



"""Module containing wrappers around rust objects and some utils"""
from pyqrlew.typing import PrivacyUnit, SyntheticData, DpEvent
from .pyqrlew import _Dataset, _Relation, _RelationWithDpEvent, Dialect, Strategy
import typing as t 
from sqlalchemy.engine import Engine

from dataclasses import dataclass
import logging
from uuid import uuid4 as generate_uuid
from typing import Optional, Tuple, List, Dict, Union
import json
from sqlalchemy import types
import sqlalchemy as sa
import typing as t


MAX_NUMERIC_RANGE = 2**50


class Dataset:
    """A Dataset is a set of SQL Tables.
    
    Examples
    ----------

    Creating a Dataset from an existing database with an sqlalchemy engine

        >>> import pyqrlew as qrl
        >>> from sqlalchemy import create_engine
        >>> engine = create_engine("postgresql+psycopg2://****/mydatabase") 
        >>> dataset = Dataset.from_database(name='extract', engine=DB.engine(), schema_name='extract', range=False, possible_values_threshold=None)

    Creating a Dataset from queries and a previous dataset. Here with WHERE statement we also determine the bounds on the age column.

        >>> queries = [
        >>>    (("ds_name", "new_schema", "tab1"), 'SELECT * FROM extract.census WHERE age < 18 AND age > 0'),
        >>>    (("ds_name", "new_schema", "tab2"), 'SELECT * FROM extract.census WHERE age >= 18  AND age < 120'),
        >>> ]
        >>> new_dataset = dataset.from_queries(queries)

    """

    CONSTRAINT_UNIQUE: str = '_UNIQUE_' 


    def __init__(self, dataset: _Dataset) -> None:
        self._dataset = dataset

    def __str__(self) -> str:
        return self._dataset.__str__()

    @staticmethod
    def from_str(dataset: str, schema: str, size: str) -> 'Dataset':
        """Factory method to create a Dataset from string representations
        compatible with protocol buffers defined `here. <https://github.com/Qrlew/sarus/tree/main/sarus_data_spec/protobuf>`_
        
        Args:
            dataset (str): string representation of a dataset
            schema (str): string representation of the dataset's schema
            size (str): string representation of the dataset's size
        """
        return Dataset(_Dataset(dataset, schema, size))

    @staticmethod
    def from_database(
        name: str,
        engine: Engine,
        schema_name: t.Optional[str]=None,
        ranges: bool=False,
        possible_values_threshold: t.Optional[int]=None
    ) -> 'Dataset':
        """Builds a `Dataset` from a sqlalchemy `Engine`.

        Args:
            name (str):
                Name of the Dataset
            engine (Engine):
                The sqlalchemy `Engine` to use
            schema_name (Optional[str], optional):
                The DB schema to use. Defaults to None.
            ranges (bool, optional):
                Use the actual min and max of the data as ranges. **This is unsafe from a privacy perspective**. Defaults to False.
                If False numeric values ranges will be considered between -2.0^50 to 2.0^50.
            possible_values_threshold (Optional[int], optional):
                Use the actual observed values as range. **This is unsafe from a privacy perspective**. Defaults to None.

        Returns:
            Dataset:
        """
        return dataset_from_database(name, engine, schema_name, ranges, possible_values_threshold)

    def __getattr__(dataset: 'Dataset', schema: str) -> 'Schema':
        return Schema(dataset, schema)

    @property
    def schema(self) -> str:
        return self._dataset.schema

    @property
    def size(self) -> t.Optional[str]:
        return self._dataset.size
    
    def with_range(self, schema_name: str, table_name: str, field_name: str, min: float, max: float) -> 'Dataset':
        """Returns a new Dataset with a defined range for a given numeric column. Check out more `here! <https://qrlew.readthedocs.io/en/latest/tutorials/getting_started.html#optionally-declare-value-ranges-and-unique-constraints>`_

        Args:
            schema_name (str): schema
            table_name (str): table
            field_name (str): column
            min (float): min range
            max (float): max range
        Returns:
            Dataset:
        """
        return Dataset(self._dataset.with_range(schema_name, table_name, field_name, min, max))
    
    def with_possible_values(self, schema_name: str, table_name: str, field_name: str, possible_values: t.Iterable[str]) -> 'Dataset':
        """Returns a new Dataset with a defined possible values for a given text column. Check out more `here! <https://qrlew.readthedocs.io/en/latest/tutorials/getting_started.html#optionally-declare-value-ranges-and-unique-constraints>`_

        Args:
            schema_name (str): schema
            table_name (str): table
            field_name (str): column
            possible_values (Iterable[str]): a sequence with wanted possible values
        Returns:
            Dataset:
        """
        return Dataset(self._dataset.with_possible_values(schema_name, table_name, field_name, possible_values))
    
    def with_constraint(self, schema_name: str, table_name: str, field_name: str, constraint: t.Optional[str]) -> 'Dataset':
        """Returns a new Dataset with a constraint on given column. Check out more `here! <https://qrlew.readthedocs.io/en/latest/tutorials/getting_started.html#optionally-declare-value-ranges-and-unique-constraints>`_
        
        Args:
            schema_name (str): schema
            table_name (str): table
            field_name (str): column
            constraint (Optional[str]):  Unique or PrimaryKey
        Returns:
            Dataset:
        """
        return Dataset(self._dataset.with_constraint(schema_name, table_name, field_name, constraint))
    
    def relations(self) -> t.Iterable[t.Tuple[t.List[str], 'Relation']]:
        """Returns the Dataset's Relations and their corresponding path"""
        return [(path, Relation(rel)) for (path, rel) in self._dataset.relations()]

    def relation(self, query: str, dialect: t.Optional['Dialect']=None) -> 'Relation':
        """Returns a Relation from am SQL query.
        
        Args:
            query (str): SQL query used to build the Relation.
            dialect (Optional[Dialect]): query's dialect. If not provided, it is assumed to be PostgreSql.
        Returns:
            Relation:
        """
        return Relation(self._dataset.relation(query, dialect))

    def from_queries(self, queries: t.Iterable[t.Tuple[t.Iterable[str], str]], dialect: t.Optional['Dialect']=None) -> 'Dataset':
        """Returns a dataset from queries.
        
        Args:
            queries (Iterable[Tuple[Iterable[str], str]]): A sequence of (path, SQL query).
                The resulting Dataset will have a Relation for each query identified in the dataset
                by the corresponding path.
            dialect (Optional[Dialect]): queries dialect. If not provided, it is assumed to be PostgreSql.
        Returns:
            Dataset:
        """
        return Dataset(self._dataset.from_queries(queries, dialect))


class Relation:
    """A Relation is a Dataset transformed by a SQL query.
    
    Example
    ----------

    Create a relation from a dataset

    >>> from pyqrlew import Dialect
    >>> query = "SELECT AVG(age), sex FROM extract.census GROUP BY sex"
    >>> relation = dataset.relation(query, Dialect.PostgreSql)

    Or alternatively 

    >>> from pyqrlew import Relation
    >>> relation = Relation.from_query(query=query, dataset=dataset, dialect=Dialect.PostgreSql) 

    """
    
    def __init__(self, relation: _Relation) -> None:
        self._relation = relation

    def __str__(self) -> str:
        return self._relation.__str__()

    @staticmethod
    def from_query(query: str, dataset: Dataset, dialect: t.Optional['Dialect']=None) -> 'Relation':
        """Builds a `Relation` from a query and a dataset

        Args:
            query (str): sql query.
            dataset (Dataset): the Dataset.
            dialect (Optional[Dialect]): query's dialect.
                If not provided, it is assumed to be PostgreSql
        
        Returns:
            Relation:
        """
        return Relation(_Relation.from_query(query, dataset._dataset, dialect))

    def to_query(self, dialect: t.Optional[Dialect]=None) -> str:
        """Returns an SQL representation of the Relation.

        Args:
            dialect (Optional[Dialect]): dialect of generated sql query. If no dialect is provided,
                the query will be in PostgreSql. 

        Returns:
            str:
        """
        return self._relation.to_query(dialect)

    def rewrite_as_privacy_unit_preserving(
        self,
        dataset: Dataset,
        privacy_unit: PrivacyUnit,
        epsilon_delta: t.Dict[str, float],
        max_multiplicity: t.Optional[float]=None,
        max_multiplicity_share: t.Optional[float]=None,
        synthetic_data: t.Optional[SyntheticData]=None,
        strategy: t.Optional[Strategy]=None,
    ) -> 'RelationWithDpEvent':
        """Returns as RelationWithDpEvent where it's relation propagates the privacy unit
        through the query. Check out more `here! <https://qrlew.readthedocs.io/en/latest/tutorials/rewrite_with_dp.html#rewritting-with-dp>`_
        
        Args:
            dataset (Dataset):
                Dataset with needed relations
            privacy_unit (PrivacyUnit):
                Definition of privacy unit to be protected. Check out more `here <https://qrlew.readthedocs.io/en/latest/tutorials/rewrite_with_dp.html#the-privacy-unit>`_
            epsilon_delta (Mapping[str, float]): epsilon and delta budget
            max_multiplicity (Optional[float]): maximum number of rows per privacy unit in absolute terms
            max_multiplicity_share (Optional[float]): maximum number of rows per privacy unit in relative terms
                w.r.t. the dataset size. The actual max_multiplicity used to bound the PU contribution will be
                minimum(max_multiplicity, max_multiplicity_share*dataset.size).
            synthetic_data (Optional[Sequence[Tuple[Sequence[str],Sequence[str]]]]): Sequence of pairs 
                of original table path and its corresponding synthetic version. Each table must be specified.
                (e.g.: (["retail_schema", "features"], ["retail_schema", "features_synthetic"])).
            strategy (Optional[Strategy]): Strategy to follow during privacy tracking. If not provided the Hard Strategy will be followed
        Returns:
            RelationWithDpEvent: 
        """
        return RelationWithDpEvent(self._relation.rewrite_as_privacy_unit_preserving(
            dataset._dataset,
            privacy_unit,
            epsilon_delta,
            max_multiplicity,
            max_multiplicity_share,
            synthetic_data,
            strategy
        ))

    def rewrite_with_differential_privacy(
        self,
        dataset: Dataset,
        privacy_unit: PrivacyUnit,
        epsilon_delta: t.Dict[str, float],
        max_multiplicity: t.Optional[float]=None,
        max_multiplicity_share: t.Optional[float]=None,
        synthetic_data: t.Optional[SyntheticData]=None,
    ) -> 'RelationWithDpEvent':
        """It transforms a Relation into its differentially private equivalent. Check out more `here! <https://qrlew.readthedocs.io/en/latest/tutorials/rewrite_with_dp.html#rewritting-with-dp>`_

        Args:
            dataset (Dataset):
                Dataset with needed relations
            privacy_unit (PrivacyUnit):
                Definition of privacy unit to be protected. Check out more `here <https://qrlew.readthedocs.io/en/latest/tutorials/rewrite_with_dp.html#the-privacy-unit>`_
            epsilon_delta (Mapping[str, float]): epsilon and delta budget
            max_multiplicity (Optional[float]): maximum number of rows per privacy unit in absolute terms
            max_multiplicity_share (Optional[float]): maximum number of rows per privacy unit in relative terms
                w.r.t. the dataset size. The actual max_multiplicity used to bound the PU contribution will be
                minimum(max_multiplicity, max_multiplicity_share*dataset.size).
            synthetic_data (Optional[Sequence[Tuple[Sequence[str],Sequence[str]]]]): Sequence of pairs 
                of original table path and its corresponding synthetic version. Each table must be specified.
                (e.g.: (["retail_schema", "features"], ["retail_schema", "features_synthetic"])).

        Returns:
            RelationWithDpEvent: 
        """
        return RelationWithDpEvent(self._relation.rewrite_with_differential_privacy(
            dataset._dataset,
            privacy_unit,
            epsilon_delta,
            max_multiplicity,
            max_multiplicity_share,
            synthetic_data
        ))

    def compose(self, relations: t.Iterable[t.Tuple[t.Iterable[str], 'Relation']]) -> 'Relation':
        """It composes itself with other relations. It substitute its Tables with the corresponding relation in relations
            with the same path. Schemas in the relations to be composed should be compatible with
            the schema of the corresponding table otherwise an error is raised.

        Args:
            relations (t.Iterable[t.Tuple[t.Iterable[str], 'Relation']]): An iterable of pairs of a path and a Relation
                Tables in self will be substituted by the relation with the matching path.
        
        Returns:
            Relation: 
        """
        converted_relations = [(path, rel._relation) for (path, rel) in relations]
        return Relation(self._relation.compose(converted_relations))

    def rename_fields(self, fields: t.Iterable[t.Tuple[str, str]])  -> 'Relation':
        """It renames fields in the Relation
        
        Args:
            fields (t.Iterable[t.Tuple[str, str]]): An iterable of pairs with
                field name to be renamed and the new name
        Returns:
            Relation:
        """
        return Relation(self._relation.rename_fields(fields))

    def with_field(self, name: str, expr: str) -> 'Relation':
        """It creates a new Relation from self with a prepended new field
        
        Args:
            name (str): The field name
            expr (str): An SQL expression following the Generic dialect
        Returns:
            Relation:
        """
        return Relation(self._relation.with_field(name, expr))

    def schema(self) -> str:
        "Returns a string representation of the Relation's schema."
        return self._relation.schema()
    
    def type(self) -> str:
        "Returns a protobuf compatible string representation of the Relation's data type."
        return self._relation.type_()

    def dot(self) -> str:
        "GraphViz representation of the `Relation`"
        return self._relation.dot()

class RelationWithDpEvent:
    """Object containing a differentially private (DP) or privacy unit
    preserving (PUP) relation and the associated DpEvent."""

    def __init__(self, relation_with_dpevent: _RelationWithDpEvent) -> None:
        self.relation_with_dpevent = relation_with_dpevent

    def relation(self) -> Relation:
        """Returns the DP or PUP relation"""
        return Relation(self.relation_with_dpevent.relation())
    
    def dp_event(self) -> DpEvent:
        """Returns the DpEvent associated with the relation."""
        return self.relation_with_dpevent.dp_event()


def dataset_from_database(
    name: str,
    engine: Engine,
    schema_name: Optional[str]=None,
    ranges: bool=False,
    possible_values_threshold: Optional[int]=None
) -> Dataset:
    """Builds a `Dataset` from a sqlalchemy `Engine`

    Args:
        name (str):
            Name of the Dataset
        engine (Engine):
            The sqlalchemy `Engine` to use
        schema_name (Optional[str], optional):
            The DB schema to use. Defaults to None.
        ranges (bool, optional):
            Use the actual min and max of the data as ranges. **This is unsafe from a privacy perspective**. Defaults to False.
        possible_values_threshold (Optional[int], optional):
            Use the actual observed values as range. **This is unsafe from a privacy perspective**. Defaults to None.

    Returns:
        Dataset:
    """

    metadata = sa.MetaData()
    metadata.reflect(engine, schema=schema_name)

    def dataset_schema_size() -> Tuple[dict, dict, Optional[dict]]:
        """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
        ds = dataset()
        return (
            ds,
            schema(ds),
            size(ds),
        )

    def dataset() -> dict:
        """Returns a Json representation compatible with [Protocol Buffers](https://protobuf.dev/reference/protobuf/google.protobuf/#google.protobuf.Struct)
        of the dataset
        """
        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Dataset',
            'uuid': generate_uuid().hex,
            'name': 'Transformed',
            'spec': {
                'transformed': {
                    'transform': generate_uuid().hex,
                    'arguments': [],
                    'named_arguments': {},
                },
            },
            'properties': {},
            'doc': 'This ia a demo dataset for testing purpose',
        }

    def schema(dataset: dict) -> dict:
        """Returns a Json representation compatible with [Protocol Buffers](https://protobuf.dev/reference/protobuf/google.protobuf/#google.protobuf.Struct)
        of the schema
        """

        tables_and_have_admin = [table(metadata.tables[name]) for name in metadata.tables]
        have_admin = [table_and_has_admin[1] for table_and_has_admin in tables_and_have_admin]

        tables = {"fields": [table_and_has_admin[0] for table_and_has_admin in tables_and_have_admin]}

        if schema_name is not None:
            tables = {
                "fields": [{
                    'name': schema_name,
                    'type': {
                        'name': 'Union',
                        'union': tables,
                        'properties': {
                            'public_fields': '[]'
                        },
                    }
                }],
            }

        sarus_is_public = {
            "name": "sarus_is_public",
            "type": {
                "boolean": {},
                "name": "Boolean",
                "properties": {}
            }
        }

        sarus_privacy_unit = {
            "name": "sarus_privacy_unit",
            "type": {
                "name": "Optional",
                "optional": {
                "type": {
                    "id": {
                    "base": "STRING",
                    "unique": False
                    },
                    "name": "Id",
                    "properties": {}
                }
                },
                "properties": {}
            }
        }

        sarus_weights = {
            "name": "sarus_weights",
            "type": {
                "float": {
                "base": "FLOAT64",
                "max": 1.7976931348623157e+308,
                "min": 0.0,
                "possible_values": []
                },
                "name": "Float64",
                "properties": {}
            }
        }
        if any(have_admin):
            return {
                '@type': 'sarus_data_spec/sarus_data_spec.Schema',
                'uuid': generate_uuid().hex,
                'dataset': dataset['uuid'],
                'name': name,
                'type': {
                    # Slugname
                    'name': name.lower(),
                    'struct': {
                        'fields': [
                            {
                                'name': 'sarus_data',
                                'type': {
                                    'name': 'Union',
                                    'union': tables,
                                    'properties': {
                                        'public_fields': '[]',
                                    },
                                },
                            },
                            sarus_is_public,
                            sarus_privacy_unit,
                            sarus_weights,
                        ],
                    },
                    'properties': {}
                },
                'privacy_unit': {
                    'label': 'sarus_data',
                    'paths': [],
                    'properties': {},
                },
                'properties': {
                    'max_max_multiplicity': '1',
                    'foreign_keys': '',
                    'primary_keys': '',
                }
            }
        else:
            return {
                '@type': 'sarus_data_spec/sarus_data_spec.Schema',
                'uuid': generate_uuid().hex,
                'dataset': dataset['uuid'],
                'name': name,
                'type': {
                    # Slugname
                    'name': name.lower(),
                    'struct': {
                        'fields': [
                            {
                                'name': 'sarus_data',
                                'type': {
                                    'name': 'Union',
                                    'union': tables,
                                    'properties': {
                                        'public_fields': '[]',
                                    },
                                },
                            }
                        ],
                    },
                    'properties': {}
                },
                'privacy_unit': {
                    'label': 'sarus_data',
                    'paths': [],
                    'properties': {},
                },
                'properties': {
                    'max_max_multiplicity': '1',
                    'foreign_keys': '',
                    'primary_keys': '',
                }
            }


    def table(tab: sa.Table) -> t.Tuple[dict, bool]:
        min_max_possible_values = compute_min_max_possible_values(tab)
        admin_cols = ['sarus_weights', 'sarus_is_public', 'sarus_privacy_unit']
        col_names = [col.name for col in tab.columns]
        has_admin = [admin_col in col_names for admin_col in admin_cols]
        return ({
            'name': tab.name,
            'type': {
                'name': 'Struct',
                'struct': {
                    'fields': [
                        column(
                            col,
                            min=t.cast(t.Optional[str], min_max_possible_values[col.name]["min"]),
                            max=t.cast(t.Optional[str], min_max_possible_values[col.name]["max"]),
                            possible_values=t.cast(t.List[str], min_max_possible_values[col.name]["possible_values"])
                        ) for col in tab.columns
                        if col.name not in admin_cols
                    ],
                },
                'properties': {},
            }
        }, all(has_admin))

    def compute_min_max_possible_values(tab: sa.Table) -> Dict[str, Dict[str, Union[t.Optional[str], List[str]]]]:
        """Send 3 SQL queries for loading the bounds"""
        values: Dict[str, Dict[str, Union[t.Optional[str], List[str]]]] = {
            col.name: {
                'min': None,
                'max': None,
                'possible_values': []
            } for col in tab.columns
        }
        intervals_types = [
            types.Integer, types.BigInteger,
            types.Float, types.Numeric,
            types.String, types.Text, types.Unicode, types.UnicodeText,
            types.Date, types.DateTime, types.Time
        ]
        interval_cols = [
            col for col in tab.columns
            if any(isinstance(col.type, t) for t in intervals_types)
        ]

        if ranges and len(interval_cols) != 0:
            min_query = sa.select(*[sa.func.cast(sa.func.min(col), sa.String).label(col.name) for col in interval_cols]).select_from(tab)
            max_query = sa.select(*[sa.func.cast(sa.func.max(col), sa.String).label(col.name) for col in interval_cols]).select_from(tab)

            with engine.connect() as conn:
                min_results = conn.execute(min_query).fetchone()
                max_results = conn.execute(max_query).fetchone()

            for col, min_val, max_val in zip(
                    interval_cols, t.cast(t.Iterable[t.Any], min_results), t.cast(t.Iterable[t.Any], max_results)
                ):
                values[col.name]['min'] = min_val
                values[col.name]['max'] = max_val

        if possible_values_threshold is not None and len(interval_cols) != 0:
            tablename = f"\"{tab.name}\"" if tab.schema is None else f"\"{tab.schema}\".\"{tab.name}\""
            values_query = sa.text(
                "SELECT " + ','.join([
                    f"CASE WHEN COUNT(DISTINCT \"{col.name}\") <= {possible_values_threshold} "
                    f"THEN array_agg(DISTINCT CAST(\"{col.name}\" AS Text)) ELSE ARRAY[]::VARCHAR[] "
                    f"END AS \"{col.name}\" "
                    for col in interval_cols
                ])
                + f"FROM {tablename}"
            ) # case very complicate to use with loop

            with engine.connect() as conn:
                values_results = conn.execute(values_query).fetchone()

            for col, possible_values in zip(interval_cols, t.cast(t.Iterable[t.Any], values_results)):
                values[col.name]['possible_values'] = [str(v) for v in possible_values]

        return values

    def column(col: sa.Column, min:Optional[str]=None, max:Optional[str]=None, possible_values:List[str]=[]) -> dict:
        if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
            min = '-9223372036854775808' if min is None else min
            max = '9223372036854775807' if max is None else max
            return {
                'name': col.name,
                'type': {
                    'name': 'Integer',
                    'integer': {
                        'base': 'INT64',
                        'min': min,
                        'max': max,
                        'possible_values': possible_values,
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
            min = f'{-MAX_NUMERIC_RANGE}' if min is None else min
            max = f'{MAX_NUMERIC_RANGE}' if max is None else max
            return {
                'name': col.name,
                'type': {
                    'name': 'Float64',
                    'float': {
                        'base': 'FLOAT64',
                        'min': min,
                        'max': max,
                        'possible_values': [],
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):

            return {
                'name': col.name,
                'type': {
                    'name': 'Text UTF-8',
                    'text': {
                        'encoding': 'UTF-8',
                        'min': min,
                        'max': max,
                        'possible_values': possible_values,
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Boolean):
            return {
                'name': col.name,
                'type': {
                    'name': 'Boolean',
                    'boolean': {},
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
            min = '01-01-01 00:00:00' if min is None else min
            max = '9999-12-31 00:00:00' if max is None else max
            return {
                'name': col.name,
                'type': {
                    'name': 'Datetime',
                    'datetime': {
                        'format': '%Y-%m-%d %H:%M:%S',
                        'min': min,
                        'max': max,
                    },
                    'properties': {},
                },
            }
        else:
            return {
                'name': col.name,
                'type': {
                    'name': 'Type',
                    'type': {},
                    'properties': {},
                },
            }

    def size(dataset: dict) -> dict:
        tables = {'fields': [table_size(metadata.tables[name]) for name in metadata.tables]}
        if schema_name is not None:
            tables = {
                'fields': [
                    {
                        'name': schema_name,
                        'statistics': {
                            'name': 'Union',
                            'union': tables,
                            'properties': {},
                        },
                        'properties': {},
                    },
                ],
            }

        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Size',
            'uuid': generate_uuid().hex,
            'dataset': dataset['uuid'],
            'name': f'{name}_sizes',
            'statistics': {
                'name': 'Union',
                'union': tables,
                'properties': {},
            },
            'properties': {},
        }

    def table_size(tab: sa.Table) -> dict:
        with engine.connect() as conn:
            result = conn.execute(sa.select(sa.func.count()).select_from(tab))
            size = t.cast(int, result.scalar())
        multiplicity = 1.0
        return {
            'name': tab.name,
            'statistics': {
                'name': 'Struct',
                'struct': {
                    'fields': [column_size(col, size, multiplicity) for col in tab.columns],
                    'size': str(size),
                    'multiplicity': multiplicity,
                },
                'properties': {},
            }
        }

    def column_size(col: sa.Column, size: int, multiplicity: float) -> dict:
        if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
            return {
                'name': col.name,
                'statistics': {
                    "integer": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Integer",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
            return {
                'name': col.name,
                'statistics': {
                    "float": {
                        "distribution": {
                            "double": {
                                "max": f'{MAX_NUMERIC_RANGE}',
                                "min": f'{-MAX_NUMERIC_RANGE}',
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Float",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):
            return {
                'name': col.name,
                'statistics': {
                    "text": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Text",
                    "properties": {}
                }
            }

        elif isinstance(col.type, types.Boolean):
            return {
                'name': col.name,
                'statistics': {
                    "boolean": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Boolean",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
            return {
                'name': col.name,
                'statistics': {
                    "datetime": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Datetime",
                    "properties": {}
                }
            }

        else: # TODO: support more types
            raise NotImplementedError(f"SQL -> Sarus Conversion not supported for {col.type} SQL type")

    # Gather protobufs
    dataset_dict, schema_dict, size_dict = dataset_schema_size()
    # Display when debugging
    logging.debug(json.dumps(dataset_dict))
    logging.debug(json.dumps(schema_dict))
    logging.debug(json.dumps(size_dict))
    # Return the result
    return Dataset.from_str(json.dumps(dataset_dict), json.dumps(schema_dict), json.dumps(size_dict))



# A method to get select a schema
def schema(dataset: Dataset, schema: str) -> 'Schema':
    return Schema(dataset, schema)


@dataclass
class Schema:
    dataset: Dataset
    schema: str

    def __getattr__(self, table: str) -> 'Table':
        return Table(self.dataset, self.schema, table)

@dataclass
class Table:
    dataset: Dataset
    schema: str
    table: str

    def __getattr__(self, column: str) -> 'Column':
        return Column(self.dataset, self.schema, self.table, column)

    def relation(self) -> Relation:
        return next(
            rel
            for (path, rel) in self.dataset.relations()
            if path[1]==self.schema and path[2]==self.table
        )

@dataclass
class Column:
    dataset: Dataset
    schema: str
    table: str
    column: str

    def with_range(self, min:float, max:float) -> Dataset:
        return self.dataset.with_range(self.schema, self.table, self.column, min, max)
    
    def with_possible_values(self, with_possible_values: List[str]) -> Dataset:
        return self.dataset.with_possible_values(self.schema, self.table, self.column, with_possible_values)
    
    def with_constraint(self, constraint: t.Optional[str]) -> Dataset:
        return self.dataset.with_constraint(self.schema, self.table, self.column, constraint)

    def with_unique_constraint(self) -> Dataset:
        return self.with_constraint(Dataset.CONSTRAINT_UNIQUE)
    
    def with_no_constraint(self) -> Dataset:
        return self.with_constraint(None)
    


import types
import sys
from pyqrlew.io.postgresql import PostgreSQL, dataset_from_database
from pyqrlew.io.sqlite import SQLite

from pyqrlew.io.utils import *

# This is for compatibility with existing notebooks where we do:
# from pyqrlew.io.dataset import dataset_from_database
dataset = types.ModuleType('pyqrlew.io.dataset')
# Populate the module with the desired function
dataset.dataset_from_database = dataset_from_database # type: ignore
# Add the module to sys.modules
sys.modules['pyqrlew.io.dataset'] = dataset



import typing as t
import pyqrlew as qrl
import qrlew_datasets.database as db  # type: ignore   #module is installed, but missing library stubs or py.typed marker 
from pyqrlew.typing import PrivacyUnit, SyntheticData


class StochasticTester:
    def __init__(
            self,
            database: db.Database,
            delta: float,
            privacy_unit: PrivacyUnit,
            synthetic_data: SyntheticData
        ) -> None:
        self.database = database
        self.delta = delta
        self.privacy_unit = privacy_unit
        self.synthetic_data = synthetic_data

    def compute_dp_relation(self, query: str, epsilon: float) -> qrl.Relation:
        budget = {"epsilon": epsilon, "delta": self.delta}
        relation = self.database.dataset().relation(query)

        relation_with_privatequeries = relation.rewrite_with_differential_privacy(
            dataset=self.database.dataset(),
            privacy_unit=self.privacy_unit,
            epsilon_delta=budget,
            synthetic_data=self.synthetic_data,
        )
        return t.cast(qrl.Relation, relation_with_privatequeries.relation())

    def utility_per_epsilon(
            self,
            query: str,
            epsilons: t.List[float],
            n_sim: int
        ) -> t.Dict[float, list]:

        print('computing utility per epsilon ...')
        relations = {
            epsilon: self.compute_dp_relation(query, epsilon)
            for epsilon in epsilons
        }
        print('done')
        return {
            epsilon: [
                self.database.eval(dp_relation)[0][0]
                for _ in range(n_sim)
            ]
            for epsilon, dp_relation in relations.items()
        }

    def compute_results(
            self,
            query: str,
            epsilon: float,
            n_sim: int
        ) -> list:
        dp_relation = self.compute_dp_relation(query, epsilon)
        return [
            self.database.eval(dp_relation)[0][0]
            for _ in range(n_sim)
        ]


import os
import numpy as np
import pandas as pd
from pyqrlew.io import PostgreSQL
from termcolor import colored
from pyqrlew import Dialect

def are_dataframes_almost_equal(df1, df2, float_tolerance=1e-6):
    # Check if the shape is the same
    if df1.shape != df2.shape:
        print("Shapes are not equal")
        return False

    # Check if column names are the same
    if not df1.columns.equals(df2.columns):
        print("Columns are not the same")
        return False

    # Replace NaN values with a placeholder for comparison
    df1 = df1.fillna('nan_placeholder')
    df2 = df2.fillna('nan_placeholder')

    # Check if values are almost equal (considering float tolerance)
    numeric_columns = df2.select_dtypes(include=[np.number]).columns
    for col in numeric_columns:
        if not np.all(np.isclose(
            df1[col].astype(float),
            df2[col].astype(float),
            rtol=float_tolerance
        )):
            print(f"In {col}, \n{list(df1[col])} \n!= \n{list(df2[col])}")
            return False

    # Check if non-numeric values are equal
    non_numeric_columns = df2.select_dtypes(exclude=[np.number]).columns
    for col in non_numeric_columns:
        if not df1[col].equals(df2[col]):
            print(f"In {col}, \n{list(df1[col])} \n!= \n{list(df2[col])}")
            print(f"type: {df1[col].dtype} {df2[col].dtype}")
            return False

    return True

def test_queries_consistency(queries):
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db

    for query in queries:
        print(f"\n{colored(query, 'blue')}")
        replaced_query = query.replace("census", "extract.census").replace("beacon", "extract.beacon")
        result = pd.read_sql(replaced_query, database.engine())            
        relation = dataset.relation(query)
        new_query = relation.to_query()
        print('===================')
        print(f"{colored(query, 'red')} -> {colored(new_query, 'green')}")
        try:
            result_after_rewriting = pd.read_sql(new_query, database.engine())
        except TypeError as e:
            print(f"Sending {new_query}\nfailed with error:\n{e}")
            return
        print(f'{result}')
        print(f'{result_after_rewriting}')
        print('===================')
        assert are_dataframes_almost_equal(result, result_after_rewriting)

def test_queries_differential_privacy(queries):
    """Test that we can rewrite the query into DP then execute it and check that the exact and DP results have the same type"""
    database = PostgreSQL()
    privacy_unit = [
        ("census", [], "_PRIVACY_UNIT_ROW_"),
        ("beacon", [], "_PRIVACY_UNIT_ROW_"),
    ]
    budget = {"epsilon": 1.0, "delta": 5e-4}
    synthetic_data = [
        (["extract", "census"], ["extract", "census"]),
        (["extract", "beacon"], ["extract", "beacon"]),
    ]
    dataset = database.extract()
    for query in queries:
        print(f"{colored(query, 'blue')}")
        relation = dataset.relation(query)
        dp_relation = relation.rewrite_with_differential_privacy(
            dataset,
            privacy_unit,
            budget,
            None,
            None,
            synthetic_data,
        ).relation()
        results = pd.read_sql(relation.to_query(), database.engine())
        dp_results = pd.read_sql(dp_relation.to_query(), database.engine())
        if len(dp_results) != 0:
            assert (results.columns == dp_results.columns).all()


def test_simple_mssql_translation():
    database = PostgreSQL()
    dataset = database.extract()
    query = "SELECT * FROM census LIMIT 10;"
    relation = dataset.relation(query, Dialect.PostgreSql)
    translated = relation.to_query(Dialect.MsSql)
    assert "SELECT TOP (10) * FROM" in translated


def test_quoting():
    database = PostgreSQL()
    dataset = database.extract()
    query = """
    SELECT 
        age AS "my age"
    FROM census;"""
    relation = dataset.relation(query, Dialect.PostgreSql)
    translated = relation.to_query(Dialect.PostgreSql)
    assert translated.replace(' ', '')=='''
        WITH "map_p5al" ("my age") AS (
            SELECT "age" AS "my age" FROM "extract"."census"
        ) SELECT * FROM "map_p5al"
    '''.replace('\n', '').replace(' ', '')
    results = pd.read_sql(relation.to_query(), database.engine())




import os
import numpy as np
import pandas as pd
from pyqrlew.io import PostgreSQL
from pyqrlew import Dialect, Dataset
from pyqrlew.utils import display_graph
from pyqrlew.wrappers import Relation

# pytest -s tests/test_dataset.py::test_ranges
def test_ranges():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    new_dataset = dataset.extract.census.age.with_range(20, 42)
    relation = new_dataset.relation('SELECT age FROM extract.census')
    print(relation.schema())
    assert(relation.schema()=='{age: int[20 42]}')

# pytest -s tests/test_dataset.py::test_possible_values
def test_possible_values():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    new_dataset = dataset.extract.census.workclass.with_possible_values(['Local-gov', 'Private'])
    relation = new_dataset.relation('SELECT workclass FROM extract.census')
    print(relation.schema())
    assert(relation.schema()=='{workclass: str{Local-gov, Private}}')

# pytest -s tests/test_dataset.py::test_constraint
def test_constraint():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    new_dataset = dataset.extract.census.age.with_unique_constraint()
    relation = new_dataset.relation('SELECT age FROM extract.census')
    print(relation.schema())
    assert(relation.schema()=='{age: int[20 90] (UNIQUE)}')

# pytest -s tests/test_dataset.py::test_relation
def test_relation():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    relation = dataset.extract.census.relation()
    print(relation.schema())


# pytest -s tests/test_dataset.py::test_from_queries
def test_from_queries():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db
    queries = [
        (("dataset_name", "my_schema", "boomers",), "SELECT * FROM extract.census WHERE age >= 60"),
        (("dataset_name", "my_schema", "genx",), "SELECT * FROM extract.census WHERE age >= 40 AND age < 60"),
        (("dataset_name", "my_schema", "millenials",), "SELECT * FROM extract.census WHERE age >= 30 AND age < 40"),
        (("dataset_name", "my_schema", "genz",), "SELECT * FROM extract.census WHERE age < 30"),
    ]

    new_ds = dataset.from_queries(queries)
    genx = new_ds.my_schema.genx.relation()
    print(genx.schema())

# pytest -s tests/test_dataset.py::test_from_dp_compiled_queries
def test_from_dp_compiled_queries():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db

    query = "SELECT COUNT(*) FROM extract.census"
    rel = Relation.from_query(query, dataset)
    privacy_unit = [
        ("census", [], "_PRIVACY_UNIT_ROW_")
    ]
    import sqlalchemy as sa
    epsilon_delta={"epsilon": 1.0, "delta": 1e-3}
    dprel_and_event = rel.rewrite_with_differential_privacy(
        dataset=dataset,
        synthetic_data=None,
        privacy_unit=privacy_unit,
        epsilon_delta=epsilon_delta,
    )
    
    dprel = dprel_and_event.relation()
    # display_graph(dprel.dot())

    new_rel = Relation.from_query(dprel.to_query(), dataset)
    # display_graph(new_rel.dot())


def test_from_database(tables, engine):
    ds =  Dataset.from_database('ds_name', engine, None)
    print(ds)
    print(ds.schema)
    paths_rels = ds.relations()
    for path, rel in paths_rels:
        print(path)
        # display_graph(rel.dot())


def test_from_str():
    dataset = """
    {
    "@type": "sarus_data_spec/sarus_data_spec.Dataset",
    "doc": "",
    "name": "Transformed",
    "properties": {},
    "spec": {
        "transformed": {
        "arguments": [
            "c07f899129e90190f48db0aebbff8abe"
        ],
        "named_arguments": {
            "privacy_unit_tracking_paths": "ece2e9d477e6c88513551c78499cbbad",
            "public_paths": "c3e688ad907078c5d834bc3a2cf4bca8"
        },
        "transform": "5c620caa2fcc37332fcd73d66372ffb5"
        }
    },
    "uuid": "50c5f0570960e686b724275e0adb39af"
    }
    """
    schema = """
    {
    "@type": "sarus_data_spec/sarus_data_spec.Schema",
    "dataset": "50c5f0570960e686b724275e0adb39af",
    "name": "demo_db_schema",
    "privacy_unit": {
        "label": "sarus_data",
        "paths": [
        {
            "label": "primary_table",
            "paths": [],
            "properties": {}
        },
        {
            "label": "secondary_table",
            "paths": [],
            "properties": {}
        },
        {
            "label": "primary_public_table",
            "paths": [],
            "properties": {}
        },
        {
            "label": "secondary_public_table",
            "paths": [],
            "properties": {}
        }
        ],
        "properties": {}
    },
    "properties": {},
    "type": {
        "name": "Struct",
        "properties": {},
        "struct": {
        "fields": [
            {
            "name": "sarus_data",
            "type": {
                "name": "Union",
                "properties": {
                "sarus_is_public": "False"
                },
                "union": {
                "fields": [
                    {
                    "name": "primary_table",
                    "type": {
                        "name": "Struct",
                        "properties": {
                        "merge_paths": "0",
                        "sarus_is_public": "False"
                        },
                        "struct": {
                        "fields": [
                            {
                            "name": "id",
                            "type": {
                                "id": {
                                "base": "INT32",
                                "unique": true
                                },
                                "name": "Id",
                                "properties": {}
                            }
                            },
                            {
                            "name": "integer",
                            "type": {
                                "integer": {
                                "base": "INT32",
                                "max": "99",
                                "min": "0",
                                "possible_values": []
                                },
                                "name": "Integer",
                                "properties": {
                                "possible_values": "[18, 79, 32, 54, 14, 81, 84, 36, 33, 24, 43, 48, 90, 99, 21, 70, 85, 51, 27, 63, 55, 49, 12, 39, 37, 44, 3, 26, 76, 65, 15, 42, 82, 4, 28, 9, 16, 62, 31, 68, 57, 87, 77, 6, 41, 22, 72, 25, 69, 7, 8, 96, 89, 38, 91, 66, 29, 5, 64, 71, 61, 20, 95, 88, 30, 50, 19, 78, 56, 0, 97, 58, 10, 98, 2, 47, 80, 35, 75, 73, 13, 83, 46, 17, 23, 74, 86, 67, 60, 1, 45, 53, 40, 11, 34, 52, 94, 59, 92, 93]",
                                "possible_values_length": "100"
                                }
                            }
                            },
                            {
                            "name": "float",
                            "type": {
                                "float": {
                                "base": "FLOAT32",
                                "max": 0.9992433190345764,
                                "min": 0.00011310506670270115,
                                "possible_values": []
                                },
                                "name": "Float64",
                                "properties": {
                                "possible_values_length": "301"
                                }
                            }
                            },
                            {
                            "name": "datetime",
                            "type": {
                                "datetime": {
                                "base": "INT64_NS",
                                "format": "%Y-%m-%d %H:%M:%S",
                                "max": "1980-01-31 18:51:02",
                                "min": "1980-01-01 01:04:49",
                                "possible_values": []
                                },
                                "name": "Datetime",
                                "properties": {
                                "possible_values_length": "301"
                                }
                            }
                            },
                            {
                            "name": "date",
                            "type": {
                                "date": {
                                "base": "INT32",
                                "format": "%Y-%m-%d",
                                "max": "2000-01-01",
                                "min": "1980-01-23",
                                "possible_values": []
                                },
                                "name": "Date",
                                "properties": {
                                "possible_values_length": "301"
                                }
                            }
                            },
                            {
                            "name": "boolean",
                            "type": {
                                "boolean": {},
                                "name": "Boolean",
                                "properties": {}
                            }
                            },
                            {
                            "name": "text",
                            "type": {
                                "name": "Text UTF-8",
                                "properties": {
                                "max_length": "8",
                                "min_length": "8",
                                "possible_values_length": "301",
                                "text_alphabet_name": "Simple",
                                "text_char_set": "[48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90]"
                                },
                                "text": {
                                "encoding": "UTF-8",
                                "possible_values": []
                                }
                            }
                            },
                            {
                            "name": "public_fk",
                            "type": {
                                "id": {
                                "base": "INT32",
                                "reference": {
                                    "label": "sarus_data",
                                    "paths": [
                                    {
                                        "label": "primary_public_table",
                                        "paths": [
                                        {
                                            "label": "id",
                                            "paths": [],
                                            "properties": {}
                                        }
                                        ],
                                        "properties": {}
                                    }
                                    ],
                                    "properties": {}
                                },
                                "unique": false
                                },
                                "name": "Id",
                                "properties": {}
                            }
                            }
                        ]
                        }
                    }
                    },
                    {
                    "name": "secondary_table",
                    "type": {
                        "name": "Struct",
                        "properties": {
                        "merge_paths": "0",
                        "sarus_is_public": "False"
                        },
                        "struct": {
                        "fields": [
                            {
                            "name": "id",
                            "type": {
                                "id": {
                                "base": "INT32",
                                "reference": {
                                    "label": "sarus_data",
                                    "paths": [
                                    {
                                        "label": "primary_table",
                                        "paths": [
                                        {
                                            "label": "id",
                                            "paths": [],
                                            "properties": {}
                                        }
                                        ],
                                        "properties": {}
                                    }
                                    ],
                                    "properties": {}
                                },
                                "unique": false
                                },
                                "name": "Id",
                                "properties": {}
                            }
                            },
                            {
                            "name": "integer",
                            "type": {
                                "integer": {
                                "base": "INT32",
                                "max": "99",
                                "min": "0",
                                "possible_values": []
                                },
                                "name": "Integer",
                                "properties": {
                                "possible_values": "[58, 72, 88, 37, 76, 26, 20, 0, 54, 93, 46, 28, 31, 3, 84, 30, 66, 17, 4, 27, 44, 94, 73, 9, 77, 69, 67, 15, 61, 56, 21, 42, 25, 16, 11, 49, 35, 90, 71, 81, 24, 53, 10, 74, 86, 62, 85, 13, 55, 48, 92, 83, 89, 59, 97, 5, 98, 50, 51, 6, 19, 96, 82, 41, 39, 7, 1, 2, 78, 79, 63, 47, 18, 43, 38, 57, 29, 8, 80, 70, 65, 68, 23, 33, 40, 60, 95, 99, 32, 87, 14, 36, 12, 34, 75, 22, 52, 64, 45, 91]",
                                "possible_values_length": "100"
                                }
                            }
                            },
                            {
                            "name": "float",
                            "type": {
                                "float": {
                                "base": "FLOAT32",
                                "max": 0.9992868304252625,
                                "min": 0.00011310506670270115,
                                "possible_values": []
                                },
                                "name": "Float64",
                                "properties": {
                                "possible_values_length": "301"
                                }
                            }
                            },
                            {
                            "name": "datetime",
                            "type": {
                                "datetime": {
                                "base": "INT64_NS",
                                "format": "%Y-%m-%d %H:%M:%S",
                                "max": "1980-01-31 23:23:22",
                                "min": "1980-01-01 01:04:49",
                                "possible_values": []
                                },
                                "name": "Datetime",
                                "properties": {
                                "possible_values_length": "301"
                                }
                            }
                            },
                            {
                            "name": "date",
                            "type": {
                                "date": {
                                "base": "INT32",
                                "format": "%Y-%m-%d",
                                "max": "2000-01-01",
                                "min": "1980-01-03",
                                "possible_values": []
                                },
                                "name": "Date",
                                "properties": {
                                "possible_values_length": "301"
                                }
                            }
                            },
                            {
                            "name": "boolean",
                            "type": {
                                "boolean": {},
                                "name": "Boolean",
                                "properties": {}
                            }
                            },
                            {
                            "name": "text",
                            "type": {
                                "name": "Text UTF-8",
                                "properties": {
                                "max_length": "8",
                                "min_length": "8",
                                "possible_values_length": "301",
                                "text_alphabet_name": "Simple",
                                "text_char_set": "[48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90]"
                                },
                                "text": {
                                "encoding": "UTF-8",
                                "possible_values": []
                                }
                            }
                            },
                            {
                            "name": "public_fk",
                            "type": {
                                "id": {
                                "base": "INT32",
                                "reference": {
                                    "label": "sarus_data",
                                    "paths": [
                                    {
                                        "label": "secondary_public_table",
                                        "paths": [
                                        {
                                            "label": "id",
                                            "paths": [],
                                            "properties": {}
                                        }
                                        ],
                                        "properties": {}
                                    }
                                    ],
                                    "properties": {}
                                },
                                "unique": false
                                },
                                "name": "Id",
                                "properties": {}
                            }
                            }
                        ]
                        }
                    }
                    },
                    {
                    "name": "primary_public_table",
                    "type": {
                        "name": "Struct",
                        "properties": {
                        "merge_paths": "1",
                        "sarus_is_public": "True"
                        },
                        "struct": {
                        "fields": [
                            {
                            "name": "id",
                            "type": {
                                "id": {
                                "base": "INT32",
                                "unique": true
                                },
                                "name": "Id",
                                "properties": {}
                            }
                            },
                            {
                            "name": "text",
                            "type": {
                                "name": "Text UTF-8",
                                "properties": {
                                "max_length": "8",
                                "min_length": "8",
                                "possible_values": "[\\"JPA2Z33I\\"]",
                                "possible_values_length": "1",
                                "text_alphabet_name": "Simple",
                                "text_char_set": "[50, 51, 65, 73, 74, 80, 90]"
                                },
                                "text": {
                                "encoding": "UTF-8",
                                "possible_values": [
                                    "JPA2Z33I"
                                ]
                                }
                            }
                            }
                        ]
                        }
                    }
                    },
                    {
                    "name": "secondary_public_table",
                    "type": {
                        "name": "Struct",
                        "properties": {
                        "merge_paths": "1",
                        "sarus_is_public": "True"
                        },
                        "struct": {
                        "fields": [
                            {
                            "name": "id",
                            "type": {
                                "id": {
                                "base": "INT32",
                                "unique": true
                                },
                                "name": "Id",
                                "properties": {}
                            }
                            },
                            {
                            "name": "text",
                            "type": {
                                "name": "Text UTF-8",
                                "properties": {
                                "max_length": "8",
                                "min_length": "8",
                                "possible_values": "[\\"JPA2Z33I\\"]",
                                "possible_values_length": "1",
                                "text_alphabet_name": "Simple",
                                "text_char_set": "[50, 51, 65, 73, 74, 80, 90]"
                                },
                                "text": {
                                "encoding": "UTF-8",
                                "possible_values": [
                                    "JPA2Z33I"
                                ]
                                }
                            }
                            }
                        ]
                        }
                    }
                    }
                ]
                }
            }
            },
            {
            "name": "sarus_is_public",
            "type": {
                "boolean": {},
                "name": "Boolean",
                "properties": {}
            }
            },
            {
            "name": "sarus_privacy_unit",
            "type": {
                "name": "Optional",
                "optional": {
                "type": {
                    "id": {
                    "base": "STRING",
                    "unique": false
                    },
                    "name": "Id",
                    "properties": {}
                }
                },
                "properties": {}
            }
            },
            {
            "name": "sarus_weights",
            "type": {
                "float": {
                "base": "FLOAT64",
                "max": 1.7976931348623157e+308,
                "min": 0.0,
                "possible_values": []
                },
                "name": "Float64",
                "properties": {}
            }
            }
        ]
        }
    },
    "uuid": "0995b7557812707dbdd97d826ef34fa5"
    }
    """

    real_ds = Dataset.from_str(dataset, schema, "")
    
    query = 'SELECT COUNT(DISTINCT sarus_privacy_unit) FROM primary_table'
    rel = Relation.from_query(query, real_ds)
    # display_graph(rel.dot())
